--- Core API only contains generic API methods found in LUI object.
-- @classmod coreapi

-- @type LUI

------------------------------------------------------
-- / SETUP AND LOCALS / --
------------------------------------------------------
local addonname, LUI = ...
local module = LUI:NewModule("API")
local db

local type, tonumber, pairs = type, tonumber, pairs
local strsub, strmatch, tostring = strsub, strmatch, tostring
local tinsert, tremove, format = tinsert, tremove, format
local min, max, math = min, max, math
local GetQuestDifficultyColor = GetQuestDifficultyColor
local GetFunctionCPUUsage = GetFunctionCPUUsage
local UnitReaction = UnitReaction

-- Constants
local MAX_AVG_ENTRIES = 10000
local MS_PER_SECOND = 1000

-- local variables
local cpuProfile = {}
local cpuAvgProfile = {}

module.defaults = {
	profile = {
		Enable = true,
	},
}

------------------------------------------------------
-- / DEV FUNCTIONS / --
------------------------------------------------------

--Function to find a wildcard inside the _G global table.
function GFind(arg, keyOnly)
	for k, v in pairs(_G) do
		if strmatch(k, arg) then
			LUI:Print(k,not keyOnly and v or nil)
		end
	end
end

function GFindValue(arg, keyOnly)
	for k, v in pairs(_G) do
		if v == arg then
			LUI:Print(k)
		end
	end
end

function GFindCTables()
	for k, v in pairs(_G) do
		if strmatch(k, "C_") and type(v) == "table" then
			LUI:Print(k,v)
		end
	end
end

function LUI:CPUProfile(func, include)
	-- Prepare a table entry for our func
	local name = tostring(func)
	if not cpuProfile[name] then
		cpuProfile[name] = 0
		cpuAvgProfile[name] = {}
	end

	--locals
	local cpuUsage = GetFunctionCPUUsage(func, include)
	local msExecTime = (cpuUsage - cpuProfile[name]) * MS_PER_SECOND
	cpuProfile[name] = cpuUsage
	tinsert(cpuAvgProfile[name], msExecTime)
	if #cpuAvgProfile[name] > MAX_AVG_ENTRIES then
		tremove(cpuAvgProfile[name], 1)
	end

	--Calculate statistics
	local sum = 0
	local vmin, vmax = math.huge, -math.huge
	for i = 1, #cpuAvgProfile[name] do
		local result = cpuAvgProfile[name][i]
		sum = sum + result
		vmin = min(vmin, result)
		vmax = max(vmax, result)
	end
	local avg = (sum / #cpuAvgProfile[name])

	LUI:Printf("%.2fms (min: %.2fms, max: %.2fms, avg: %.2fms)", msExecTime, vmin, vmax, avg)
end

--takes table, second arg for recursion. Prints an entire table to default chat.
function LUI:PrintFullTable(tbl,msg, recurse)
	if type(tbl) ~= "table" then return LUI:Print("Tried to Print a nil table.") end
	if not recurse then LUI:Print("-------------------------") end
	msg = msg or ""
	for k,v in pairs(tbl) do
		if type(v) == "table" then
			LUI:Print(msg,k,v)
			LUI:PrintFullTable(v,msg.."-- ", true)
		else LUI:Print(msg,k,v) end
	end
	if not recurse then LUI:Print("-------------------------") end
end

--- Dev Functions
-- @section stringfunc

--- Print a table to the chat frame
-- @param tbl Table to print.
function LUI:PrintTable(tbl)
	if type(tbl) ~= "table" then return LUI:Print("Tried to Print a nil table.") end
	LUI:Print("-------------------------")
	for k, v in pairs(tbl) do
		LUI:Print(k,v)
	end
	LUI:Print("-------------------------")
end

--Function to add a bright border around a given frame to help seeing it and its size.
function LUI:HighlightBorder(frame)
	local glowBackdrop = {
		bgFile="Interface\\Tooltips\\UI-Tooltip-Background",
		edgeFile="Interface\\AddOns\\LUI\\media\\statusbar\\glowTex.tga",
		--tile=0, tileSize=0,
		edgeSize=5,
		insets={left=3, right=3, top=3, bottom=3}
	}
	frame:SetBackdrop(glowBackdrop)
	frame:SetBackdropColor(0,0,0,0)
	frame:SetBackdropBorderColor(1,1,0,1)
end

------------------------------------------------------
-- / GENERAL API / --
------------------------------------------------------

--Count the number of entries in a table. This is done because #Table only returns array.
function LUI:Count(t,isPrint)
	local count = 0
	if type(t) == "table" then
		for _ in pairs(t) do count = count + 1 end
	end
	if isPrint then
		LUI:Print(count)
	else
		return count
	end
end

--Give us a sorted table to work with, fill the array with the keys, then sort based on the values in original table
--then we can just use a for loop to get a sorted result and call original[ sorted[i] ] for the value
--Went with a return-less approach that you need to provide the sort table because otherwise,
--I would need to create a new table every single call, and that would create needless garbage.
function LUI:SortTable(sortT, origT, sortFunc)
	wipe(sortT)
	for k in pairs(origT) do sortT[#sortT+1] = k end
	sort(sortT, sortFunc)
end

--Copy a table recursively.
function LUI:CopyTable(source, target)
	if type(target) ~= "table" then target = {} end
	for k, v in pairs(source) do
		if type(v) == "table" then
			target[k] = LUI:CopyTable(v, target[k])
		elseif not target[k] then
			target[k] = v
		end
	end
	return target
end

------------------------------------------------------
-- / VARIOUS COLOR API / --
------------------------------------------------------

-- Return r, g, b for selected faction, otherwise return white.
function LUI:GetFactionColor(faction)
	if LUI:Color(faction) then
		return LUI:Color(faction)
	else
		return 1, 1, 1
	end
end

-- Return r, g, b based on reaction of unit towards another unit.
-- If a second unit isnt given, assume player.
function LUI:GetReactionColor(unit, otherUnit)
	local reaction = UnitReaction(unit, otherUnit or "player")
	if LUI:Color(LUI.REACTION_NAMES[reaction]) then
		return LUI:Color(LUI.REACTION_NAMES[reaction])
	else
		return 1, 1, 1
	end
end

-- Return r, g, b based on level difference.
function LUI:GetDifficultyColor(level)
	local color = GetQuestDifficultyColor(level)
	return color.r, color.g, color.b
end

--Function wrappers for Good/Bad colors convenience.
function LUI:PositiveColor()
	return LUI:Color("Good")
end
function LUI:NegativeColor()
	return LUI:Color("Bad")
end

-- Based on Wowpedia's ColorGradient. Use our three gradient colors to make a color based on a percentage
-- TODO: Possibly rename some variables inside. (such as relperc.)
function LUI:ColorGradient(perc)
	if perc >= 1 then
		return LUI:PositiveColor()
	elseif perc <= 0 then
		return LUI:NegativeColor()
	end

	local segment, relperc = math.modf(perc * 2)
	local r1, r2, g1, g2, b1, b2
	if segment == 0 then
		r1, g1, b1 = LUI:Color("Bad")
		r2, g2, b2 = LUI:Color("Medium")
	elseif segment == 1 then
		r1, g1, b1 = LUI:Color("Medium")
		r2, g2, b2 = LUI:Color("Good")
	end

	local r = r1 + (r2 - r1) * relperc
	local g = g1 + (g2 - g1) * relperc
	local b = b1 + (b2 - b1) * relperc
	return r, g, b
end

--Wrapper for ColorGradient's that inverse the percent given.
function LUI:InverseGradient(perc)
	return LUI:ColorGradient(1 - perc)
end