--- Modules api contains all the generic embeddable api that modules can use to easily acess or do stuff.
-- @classmod modapi

------------------------------------------------------
-- / SETUP AND LOCALS / --
------------------------------------------------------

-- Addon building reference.
local addonname, LUI = ...
local module = LUI:GetModule("API")
local element = module:NewModule("Modules")
local Media = LibStub("LibSharedMedia-3.0")
local db

--local copies
local pairs = pairs

--Local variables
local modapi = {}
local modcreate = {}
local modColor  -- Will hold color module.

function LUI:EmbedModule(target)
	for k, v in pairs(modapi) do
		target[k] = v
	end
end

------------------------------------------------------
-- / MODULE API  / --
------------------------------------------------------

--- Module API
-- @section moduleapi

--- Fetch a color.
-- @string color Name of the color to fetch as named in the database.
-- This will check the specific module's database and then check the generic Color module.
-- This function will be really helpful later on when we add choice to use a class or theme color instead of individual.
-- @treturn number r Red
-- @treturn number g Green
-- @treturn number b Blue
-- @treturn ?number a Alpha
function modapi:Color(colorName)
	--TODO: Fix the issue with RGB colors as RGBA colors in the options
	--TODO: Add Better Element/Module support, order to check should be the element,then parent module, then Colors.
	local color
	local db = self:GetDB("profile", "Colors")
	if db and db[colorName] then
		color = db[colorName]
		-- TODO: Check for Type
	else
		color = modColor.db.profile.Colors[colorName]
	end
	if color then return color.r, color.g, color.b end
end

function modapi:AlphaColor(colorName)
	local color
	local db = self:GetDB("profile", "Colors")
	if db and db[colorName] then
		color = db[colorName]
		-- TODO: Check for Type
	else
		color = modColor.db.profile.Colors[colorName]
	end
	if color then return color.r, color.g, color.b, color.a or 1 end
end

--- Fetch a color table.
-- Same thing as Color() except it returns a table instead of RGB values.
-- @string color name of the color to fetch as named in the database.
-- @treturn table Color table.
function modapi:ColorTable(colorName)
	local color
	local db = self:GetDB("profile", "Colors")
	if db and db[colorName] then
		color = db[colorName]
		-- TODO: Check for Type
	else
		color = modColor.db.profile.Colors[colorName]
	end
	return color
end

--- Quickly Setup a FontString widget
-- @tparam table frame Frame to attach the font string to.
-- @string gName Global name for the new font stirng
-- @string mFont Name of the font to fetch as named in the database.
-- @tparam[opt] ?string layer Layer on which to create the font string, default to ARTWORK.
-- @tparam[opt] ?string hJustify Set the font horizontal alignment.
-- @tparam[opt] ?string vJustify Set the font vertical alignment.
function modapi:SetFontString(frame, gName, mFont, layer, hJustify, vJustify)
	local fs = frame:CreateFontString(gName, layer)
	local db = self:GetDB("profile", "Fonts")
	local font = db[mFont]
	fs:SetFont(Media:Fetch("font", font.Name), font.Size, font.Flag)
	if hJustify then fs:SetJustifyH(hJustify) end
	if vJustify then fs:SetJustifyV(vJustify) end
	return fs
end

function modapi:RefreshFontString(fs, mFont)
	local db = self:GetDB()
	local font = db.Fonts[mFont]
	fs:SetFont(Media:Fetch("font", font.Name), font.Size, font.Flag)
	fs:SetTextColor(self:Color(mFont))
end

--- Returns a database table. It will return the profile scope by defaults.
-- @tparam[opt] ?string scope The scope to look up. Can be one of the nine database types as specified by AceDB.
-- @tparam[opt] ?string extra The name of a table in the database to return.
function modapi:GetDB(scope, extra)
	local scope = scope or "profile"
	if self:IsElement() then
		local _, parent = self:GetParent()
		db = parent.db[scope][self:GetName()]
	else
		db = self.db[scope]
	end
	return (extra) and db[extra] or db
end

--- Check if the module is an element.
-- @treturn bool Returns true if it's an element, false if it's a module (or the LUI object)
function modapi:IsElement()
	if not self.GetParent then return false end
	return (self:GetParent() ~= "LUI") and true or false
end

--- Print exclusively for Module Messages.
-- Those prints will not appear if ModuleMessages is disabled
-- @string msg Message to print. Note that this print automatically adds the module name at the start.
function modapi:ModPrint(msg)
	local db = LUI:GetDB()
	if db.General.ModuleMessages then
		LUI:Printf("%s: %s", self:GetName(), msg)
	end
end

------------------------------------------------------
-- / MODULE CREATION-EMBEDDED API / --
------------------------------------------------------
-- @local here
function modcreate:NewElement(name, ...)
	local element = self:NewModule(name, ...)
	LUI:EmbedModule(element)
	return element
end

--Make sure every element also has a :GetParent()
function modcreate:OnModuleCreated(element)
	element.GetParent = function()
		return self:GetName(), self
	end
end

--- Get the module's parent.
-- This function returns nil if called by LUI.
-- @function modapi:GetParent
-- @treturn string The parent's name.
-- @treturn table The parent's table

--- Toggle a module's enabled state.
-- This function is not avaible to elements or LUI.
-- @function Toggle

function LUI:OnModuleCreated(module)
	module.GetParent = function()
		return self:GetName(), self
	end
	
	--Only modules with an enableButton should be toggle-able.
	module.Toggle = function()
		local name = module:GetName()
		local state = not module:IsEnabled()
		if state then
			LUI:EnableModule(name)
		else
			LUI:DisableModule(name)
		end
		local db = LUI:GetDB()
		db.modules[name] = state
	end
	
	for k, v in pairs(modcreate) do
		module[k] = v
	end
end

------------------------------------------------------
-- / FRAMEWORK FUNCTIONS / --
------------------------------------------------------
-- @local here
function element:OnInitialize()
	modColor = LUI:GetModule("Colors")
end