--[[ File: Bags\Container.lua, contains core of bags module. 

This file contains the basic of the Bags module, and the Container prototype, that will be used to make bag frames. 
Keep as geneneric as possible to allow possible expansions. (Guild Bank, Void Storage)

Container Members: 
- type - Type of container, ie: Bags, Bank
- name - Frame name for the container
- idList - Table containing a list of Bag ID for the container
- numID - The number of Bag ID for the container. Since the idList is a hash, we cant use #idList
- itemList - Contains the individual frames for the item slots, separated by ID.
- bagsBarList- Contains the individual frames for the BagsBar
- bagList - Contains the individual parent frames for the IDs.
--]]
------------------------------------------------------
-- / SETUP AND LOCALS / --
------------------------------------------------------
local addonname, LUI = ...
local module = LUI:NewModule("Bags", "AceHook-3.0")
local Media = LibStub("LibSharedMedia-3.0")
local L = LUI.L
local db

local format, pairs = format, pairs
local GetContainerNumFreeSlots = GetContainerNumFreeSlots
local CreateFrame = CreateFrame

-- Constants
local BACKPACK_TOOLTIP = BACKPACK_TOOLTIP

local BUTTON_SLOT_TEMPLATE = "ItemButtonTemplate"
local BAG_UPDATE_TIME = 0.05
local BAG_TEXTURE_SIZE = 36
local LAYOUT_OFFSET = 26

-- Local variables
local Container = {}  -- Prototype
local containerStorage = {}


-- Defaults
module.defaults = {
	profile = {
		--Container Settings
		Lock = false,
		RowSize = 16,
		Padding = 8,
		Spacing = 4,
		Scale = 1.5,
		BagBar = true,
		ItemQuality = true,
		BagNewline = false,
		ShowNew = false,
		ShowQuest = true,
		BackgroundTexture = "Blizzard Tooltip",
		BorderTexture = "Stripped_medium",
		BorderSize = 5,
		-- Keep track of frame positions
		Position = {
			["*"] = {
				X = 0,
				Y = 0,
			},
		},
		Textures = {
			Background = "Blizzard Tooltip",
			Border = "Stripped_medium",
			BorderSize = 5,
		},
		-- Fonts and Colors
		Fonts = {
			Bags = { Name = "AvantGarge_LT_Medium", Size = 12, } --Flag = "OUTLINE", },
		},
		Colors = {
			Search =      { r = 0.6,  g = 0.6,  b = 1,    a = 1,   t = "Class",      },
			Border =      { r = 0.2,  g = 0.2,  b = 0.2,  a = 1,   t = "Individual", },
			Background =  { r = 0.18, g = 0.18, b = 0.18, a = 0.8, t = "Individual", },
			Professions = { r = 0.1,  g = 0.5,  b = 0.2,                             },
			--TODO: Add support for FrameBorder and FrameBackground
			--FrameBackground = { r = 0.09, g = 0.09, b = 0.09, a = 0.8, t = "Individual", },
		},
	},
}

------------------------------------------------------
-- / CONTAINER API / --
------------------------------------------------------
function Container:Open()
	self:Show()
end

function Container:Close()
	self:Hide()
end

function Container:Toggle()
	if self:IsShown() then
		self:Hide()
	else
		self:Show()
	end
end

function Container:StartMovingFrame()
	if not db.Lock then
		self:StartMoving()
	end
end

function Container:StopMovingFrame()
	self:StopMovingOrSizing()
	local x, y = self:GetCenter()
	db.Position[self.name].X = format("%.1f", x)
	db.Position[self.name].Y = format("%.1f", y)
end

function Container:OnShow()
	self.updateBucket = self:RegisterBucketEvent("BAG_UPDATE", BAG_UPDATE_TIME, "BagUpdateEvent")
	module:RegisterEvent("ITEM_LOCK_CHANGED", function(...) self:ItemLockUpdate(...) end)
	self:Layout()
end

function Container:OnHide()
	self:UnregisterBucket(self.updateBucket)
	module:UnregisterEvent("ITEM_LOCK_CHANGED")
end

-- Function to get a DB option value, will fetch from generic, unless frame-specific settings is enabled
function Container:GetOption(name)
	--TODO: Add support for frame-specific settings
	return db[name]
end

function Container:IsValidID(id)
	return tContains(self.BAG_ID_LIST, id) and true or false
end

function Container:Layout()
	for i = 1, self.NUM_BAG_IDS do
		local id = self.BAG_ID_LIST[i]
		local itemList = self.itemList[id]
		--get a new bagCount in case bags changed.
		local bagCount = GetContainerNumSlots(id)
		if bagCount > 0 then
			self.bagList[id]:Show()
			for i = 1, bagCount do
				-- The item slots will be anchored later on.
				itemList[i] = self:NewItemSlot(id, i)
				self:SlotUpdate(itemList[i])
			end
		end
		
		--If there are more itemSlots than bagCount, hide them.
		--This way, we can reuse frames, instead of creating new ones
		for i = bagCount + 1, #itemList do
			if itemList[i] then
				itemList[i]:Hide()
			end
		end
	end
	
	self:SetAnchors()
end

function Container:SetBagsProperties()
	--local bagsBar = self.BagsBar
	
	-- Set Position
	local position = db.Position[self.name] or {}
	self:SetPoint("CENTER", UIParent, "BOTTOMLEFT", position.X or 0, position.Y or 0)
	
	-- Scale
	self:SetScale(self:GetOption("Scale"))
	
	-- Backdrops
	module:ApplyBackdrop(self.background, module.bagBackdrop)
	for i = 1, self.NUM_BAG_IDS do
		local id = self.BAG_ID_LIST[i]
		for i = 1, #self.itemList[id] do
			module:ApplyBackdrop(self.itemList[id][i], module.itemBackdrop)
		end
	end
	
	-- TODO: Integrate SetBagFonts into this
end

function Container:NewBagInfo(id)
	-- If the ID has already been created, return it
	if self.bagList[id] then return self.bagList[id] end
	
	--Create the frame
	local bagFrame = CreateFrame("Frame", self:GetName()..id, self)
	bagFrame:SetID(id)
	
	return bagFrame
end

-- Due to using Blizzard's template itemSlots which contains differences in names and format,
-- This should be handled by elements personally [using :NewItemSlot(id, slot)]
-- The elements should then call self:SetItemSlotProperties for post-creation shared code.
function Container:SetItemSlotProperties(itemSlot)
	--Make it easy to fetch Cooldown information
	itemSlot.cooldown = _G[itemSlot:GetName() .. "Cooldown"]
	itemSlot.cooldown:Show()
	
	--Update backdrop
	module:ApplyBackdrop(itemSlot, module.itemBackdrop)
end

function Container:SlotUpdate(itemSlot)
	local id, slot = itemSlot.id, itemSlot.slot
	local texture, count, locked, quality = GetContainerItemInfo(id, slot)
	
	-- Default Border when items are locked.
	if not itemSlot.lock then
		itemSlot:SetBackdropBorderColor(module:AlphaColor("Border"))
		-- Check for Profession Bag
		if module:IsProfessionBag(id) then
			itemSlot:SetBackdropBorderColor(module:AlphaColor("Professions"))
		end
	end
	
	-- Add the cooldown to our item slots.
	if itemSlot.cooldown then
		local start, duration, enable = GetContainerItemCooldown(id, slot)
		CooldownFrame_SetTimer(itemSlot.cooldown, start, duration, enable)
	end
	
	-- New item code from Blizzard's ContainerFrame.lua
	local newItemTexture = itemSlot.NewItemTexture
	local battlePayTexture = itemSlot.BattlepayItemTexture
	local flashAnim = itemSlot.flashAnim
	local newItemAnim = itemSlot.newitemglowAnim
	-- Not all item slots have a newItemTexture
	if newItemTexture then
		if self:GetOption("ShowNew") and C_NewItems.IsNewItem(id, slot) then
			if IsBattlePayItem(id, slot) then
				newItemTexture:Hide()
				battlePayTexture:Show()
			else
				if quality and NEW_ITEM_ATLAS_BY_QUALITY[quality] then
					newItemTexture:SetAtlas(NEW_ITEM_ATLAS_BY_QUALITY[quality])
				else
					newItemTexture:SetAtlas("bags-glow-white")
				end
				newItemTexture:Show()
				battlePayTexture:Hide()
			end
			if not flashAnim:IsPlaying() and not newItemAnim:IsPlaying() then
				flashAnim:Play()
				newItemAnim:Play()
			end
		else
			-- If the item is not new, hide all related textures
			newItemTexture:Hide()
			battlePayTexture:Hide()
			if flashAnim:IsPlaying() or newItemAnim:IsPlaying() then
				flashAnim:Stop()
				newItemAnim:Stop()
			end
		end
		--Make sure that the textures are the same size as the itemframe.
		battlePayTexture:SetSize(itemSlot:GetSize())
		newItemTexture:SetSize(itemSlot:GetSize())
	end
	
	-- Quest Item code from Blizzard's ContainerFrame.lua
	local questTexture = _G[itemSlot:GetName().."IconQuestTexture"]
	if questTexture then
		questTexture:SetSize(itemSlot:GetSize())
		local isQuestItem, questId, isActive = GetContainerItemQuestInfo(id, slot)
		if questId and not isActive and self:GetOption("ShowQuest") then
			questTexture:SetTexture(TEXTURE_ITEM_QUEST_BANG)
			questTexture:Show()
		elseif (questId or isQuestItem) and self:GetOption("ShowQuest") then
			questTexture:SetTexture(TEXTURE_ITEM_QUEST_BORDER)
			questTexture:Show()
		else
			questTexture:Hide()
		end
	end
	
	-- Color Border according to quality
	local itemLink = GetContainerItemLink(id, slot)
	if itemLink then
		-- Do not use earlier quality var, GetContainerInfo returns inacurate information for unusable items.
		local name, _, quality = GetItemInfo(itemLink)
		if self:GetOption("ItemQuality") and quality and quality > 1 and not itemSlot.lock then
			itemSlot:SetBackdropBorderColor(GetItemQualityColor(quality))
		end
		-- Keep the name for easy searching
		itemSlot.name = name
		itemSlot.quality = quality
	else
		-- Make sure to not keep name/quality info from previous item
		itemSlot.name = nil
		itemSlot.quality = nil
	end
	
	SetItemButtonTexture(itemSlot, texture)
	SetItemButtonCount(itemSlot, count)
	SetItemButtonDesaturated(itemSlot, locked, 0.5, 0.5, 0.5)
	
	itemSlot:Show()
end

function Container:ItemLockUpdate(event, id, slot)
	if not slot or not self:IsValidID(id) or not self.itemList[id][slot] then 
		return
	end
	self:SlotUpdate(self.itemList[id][slot])
end

function Container:BagUpdateEvent(idList)
	if not self.itemList then return end
	
	-- HACK: Whenever a bag is changed, BAG_UPDATE 0 triggers, instead of the missing bag ID.
	-- We need to reload the frame if a 0 happens, check it individually to skip the loop.
	if idList[0] then
		self:Layout()
		return
	end
	
	for id in pairs(idList) do
		if self:IsValidID(id) then
			--???: This conditional never triggers, as it should, why keep it?
			if not self.itemList[id] then
				LUI:Print("Cannot find ItemList["..id.."]. Reloading")
				self:Layout()
				return
			end
			for i = 1, #self.itemList[id] do
				self:SlotUpdate(self.itemList[id][i])
			end
		end
	end
end

-- This function will set all itemslot anchors and the container's dimensions based on that.
function Container:SetAnchors()
	-- index will help us stay positioned to prevent going above RowSize
	-- lineAnchor is going to store the first frame of every line, allowing to make a new line easily
	-- rightAnchor is going to store the rightmost frame, to set the width of the container
	-- rightIndex will be used to denote the position of the rightAnchor, to make sure it stays the rightmost frame
	-- previousAnchor is going to store the frame we just processed, so easily anchor the next one (unless newline)
	local lineAnchor, rightAnchor, previousAnchor
	local index = 0
	local rightIndex = 0
	local padding = self:GetOption("Padding")
	local spacing = self:GetOption("Spacing")
	local rowSize = self:GetOption("RowSize")
	
	for i = 1, self.NUM_BAG_IDS do
		local id = self.BAG_ID_LIST[i]
		--TODO: Add Option to newline on new bag
		-- if NewLineOnNewBag then index = 0 end
		if self:GetOption("BagNewline") then 
			index = 0
		end
		for i = 1, #self.itemList[id] do
			local itemSlot = self.itemList[id][i]
			-- Make sure to clear points to prevent errors.
			itemSlot:ClearAllPoints()
			-- ItemSlots beyond bagCount are hidden, so we don't count them
			if itemSlot:IsShown() then
				-- Increment the index for positioning
				index = index + 1
				-- if lineAnchor is nil, then its the first slot.
				if not lineAnchor then
					local xOffset = padding
					--TODO: Possibly rename LAYOUT_OFFSET to TITLEBAR_HEIGHT unless used for something else
					local yOffset = LAYOUT_OFFSET + padding
					itemSlot:SetPoint("TOPLEFT", self, "TOPLEFT", xOffset, -yOffset)
					-- Set the itemSlot to be the anchor for future slots.
					lineAnchor = itemSlot
					rightAnchor = itemSlot
					previousAnchor = itemSlot
					rightIndex = index
				-- Check to see if we need to do a newline
				elseif index == 1 or index > rowSize then
					-- The previous lineAnchor takes care of the xOffset
					local yOffset = spacing
					itemSlot:SetPoint("TOP", lineAnchor, "BOTTOM", 0, -yOffset)
					-- Since it was a newline, it becomes the new lineAnchor
					lineAnchor = itemSlot
					previousAnchor = itemSlot
					index = 1
				-- In any other situation, just anchor it to the right of the previous slot
				else
					local xOffset = spacing
					-- The previousAnchor takes care of the yOffset
					itemSlot:SetPoint("LEFT", previousAnchor, "RIGHT", xOffset, 0)
					previousAnchor = itemSlot
					-- Check to see if it becomes the new rightAnchor
					if index > rightIndex then
						rightAnchor = itemSlot
						rightIndex = index
					end
				end
			end
		end  -- end of itemList loop for current ID
	end -- end of itemList for the last ID
	
	-- Set anchors of the background frame to cover all the items. 
	self.background:SetPoint("LEFT", lineAnchor, "LEFT", -padding, 0)
	self.background:SetPoint("RIGHT", rightAnchor, "RIGHT", padding, 0)
	self.background:SetPoint("BOTTOM", lineAnchor, "BOTTOM", 0, -padding)
	self.background:SetPoint("TOP", rightAnchor, "TOP", 0, LAYOUT_OFFSET + padding)
	-- Then set the size of the container frame to be equal to the background. 
	self:SetSize(self.background:GetWidth(), self.background:GetHeight())
end

------------------------------------------------------
-- / CONTAINER: BAGBAR FUNCTIONS / --
------------------------------------------------------
function Container:CreateToolBar(name)
	local toolBar = CreateFrame("Frame", nil, self)
	toolBar:SetClampedToScreen(true)
	toolBar:SetSize(1,1)
	
	
	local bgFrame = CreateFrame("Frame", nil, toolBar)
	--Force it to the lowest frame level to prevent layering issues
	bgFrame:SetFrameLevel(toolBar:GetParent():GetFrameLevel())
	bgFrame:SetClampedToScreen(true)
	
	module:ApplyBackdrop(bgFrame, module.bagBackdrop)
	
	toolBar.slotList = {}
	toolBar.background = bgFrame
	self.toolbars[name] = toolBar
	if not self[name] then self[name] = toolBar end
	
	--Default function for toolbars to anchor slots.
	toolBar.SetAnchors = function()
		local padding = self:GetOption("Padding")
		local spacing = self:GetOption("Spacing")
		local previousAnchor, firstAnchor
		for i = 1, #toolBar.slotList do
			local slot = toolBar.slotList[i]
			slot:ClearAllPoints()
			
			if not previousAnchor then -- first slot
				slot:SetPoint("TOPLEFT", toolBar, "TOPLEFT", padding, -padding)
				previousAnchor = slot
				firstAnchor = slot
			else
				slot:SetPoint("LEFT", previousAnchor, "RIGHT", spacing, 0)
				previousAnchor = slot
			end
		end
		
		bgFrame:SetPoint("LEFT", firstAnchor, "LEFT", -padding, 0)
		bgFrame:SetPoint("TOP", firstAnchor, "TOP", 0, padding)
		bgFrame:SetPoint("BOTTOM", firstAnchor, "BOTTOM", 0, -padding)
		bgFrame:SetPoint("RIGHT", previousAnchor, "RIGHT", padding, 0)
		
		toolBar:SetSize(bgFrame:GetWidth(), bgFrame:GetHeight())
		toolBar:Show()
	end
	
	--Simple function to add a new button to the toolbar.
	toolBar.nextIndex = 1
	toolBar.AddNewButton = function(self, newButton)
		toolBar.slotList[toolBar.nextIndex] = newButton
		toolBar.nextIndex = toolBar.nextIndex + 1
	end
end

------------------------------------------------------
-- / MODULE FUNCTIONS / --
------------------------------------------------------
-- Function to set backdrop and its colors
function module:ApplyBackdrop(frame, backdrop)
	frame:SetBackdrop(backdrop)
	frame:SetBackdropColor(module:AlphaColor("Background"))
	frame:SetBackdropBorderColor(module:AlphaColor("Border"))
end

-- Funciton to create a blank slot used for tool bars, items, etc.
-- Template is optional and defaults to "ItemButtonTemplate" if missing.
function module:CreateSlot(name, parent, template)
	local button = CreateFrame("Button", name, parent, template or BUTTON_SLOT_TEMPLATE)
	button:SetSize(BAG_TEXTURE_SIZE, BAG_TEXTURE_SIZE)
	button:SetPushedTexture("")
	button:SetNormalTexture("")
	
	local normalTex = _G[name.."NormalTexture"]
	if normalTex then
		normalTex:SetSize(1,1)
	end
	
	--Make IconTexture not clash with our backdrop
	local iconTex = _G[name.."IconTexture"]
	SetItemButtonTexture(button)
	if iconTex then
		-- This removes the white/silver border found around many IconTextures
		iconTex:SetTexCoord(0.08, 0.92, 0.08, 0.92)
		iconTex:SetPoint("TOPLEFT", button, 3, -3)
		iconTex:SetPoint("BOTTOMRIGHT", button, -3, 3)
		-- This prevent the IconTextures from appearing (partially) above our itemSlot backdrop 
		iconTex:SetDrawLayer("BORDER", -1)
		iconTex:Show()
	end
	
	return button
end

function module:CreateNewContainer(name, obj)
	if containerStorage[name] then return end
	
	-- Create the frame and set properties
	local frame = CreateFrame("Frame", "LUI"..name, UIParent)
	frame:SetFrameStrata("HIGH")
	frame:EnableMouse(true)
	frame:SetMovable(true)
	frame:SetToplevel(true)
	frame:SetClampedToScreen(true)
	frame:SetSize(1,1)
	
	-- Create Background frame
	local bgFrame = CreateFrame("Frame", nil, frame)
	--When Bags and Bank are opened at the same time, there is overlap happening. FIgure a better way to fix it. 
	bgFrame:SetFrameLevel(frame:GetParent():GetFrameLevel()+1)
	bgFrame:SetClampedToScreen(true)
	frame.background = bgFrame
	
	-- Add Close Button
	local closeBtn = CreateFrame("Button", nil, frame, "UIPanelCloseButton")
	closeBtn:SetSize(32,32)
	closeBtn:SetPoint("TOPRIGHT", -3, -3)
	closeBtn:RegisterForClicks("AnyUp")
	closeBtn:SetScript("OnClick", function() frame:Close() end)
	frame.closeButton = closeBtn
	
	-- Add reference tables
	frame.toolbars = {} -- Used to store BagBar and such
	
	-- More Robust Embedding system. It will embed things from the given object.
	-- Then Container contains the shared code, hook it to existing functions or embed it. 
	for k, v in pairs(obj) do
		frame[k] = v
	end
	for k, v in pairs(Container) do
		if frame[k] then
			module:Hook(frame, k, Container[k])
		else
			frame[k] = v
		end
	end
	--Add AceBucket to the Container to process bag updates.
	LibStub("AceBucket-3.0"):Embed(frame)
	
	--Set up scripts
	frame:SetScript("OnShow", frame.OnShow)
	frame:SetScript("OnHide", frame.OnHide)
	frame:SetScript("OnMouseDown", frame.StartMovingFrame)
	frame:SetScript("OnMouseUp", frame.StopMovingFrame)
	
	
	-- Craete Search Box
	module:CreateSearchEditBox(frame)  -- Placeholder
	
	-- Create the Bag Bar
	if frame.CreateBagBar then
		frame:CreateToolBar("bagsBar")
		frame.bagsBar:SetPoint("BOTTOMLEFT", frame, "TOPLEFT", 0, 2)
		frame:CreateBagBar()
	end
	
	-- Create the Utility Bar
	if frame.CreateUtilBar then
		frame:CreateToolBar("utilBar")
		if frame.bagsBar then
			frame.utilBar:SetPoint("LEFT", frame.bagsBar, "RIGHT", 4, 0)
		else
			frame.utilBar:SetPoint("BOTTOMLEFT", frame, "TOPLEFT", 0, 2)
		end
		frame:CreateUtilBar()
	end
	
	--Preliminary table creation.
	frame.bagList = {}
	frame.itemList = {}
	for i = 1, frame.NUM_BAG_IDS do
		local id = frame.BAG_ID_LIST[i]
		frame.bagList[id] = frame:NewBagInfo(id)
		frame.itemList[id] = {}
	end
	
	--Do those needs to be here?
	--SetBagFonts -- Nope, this one seems very specific to some frames, could probably set 
	frame:SetBagsProperties()
	containerStorage[name] = frame
	--SetBagsDimensions
	
	frame:Hide()
end

-- This is a placeholder function until we get search.lua up
-- It only makes the text appear, no actual functionality.
function module:CreateSearchEditBox(parent)
	--Search Text
	local search = parent:CreateFontString(nil, "OVERLAY", "GameFonthighlightLarge")
	search:SetPoint("TOPLEFT", parent, db.Padding, -10)
	search:SetPoint("TOPRIGHT", -40, 0)
	search:SetJustifyH("LEFT")
	search:SetText(LUI:ColorToString(SEARCH, module:Color("Search")))
end

function module:IsProfessionBag(id)
	local _, bagType = GetContainerNumFreeSlots(id)
	if bagType and bagType > 0 then
		return true
	end
	return false
end

------------------------------------------------------
-- / OPTIONS FUNCTIONS / --
------------------------------------------------------

--Set the scale of all containers currently shown
function module:SetContainerScale()
	for name, container in pairs(containerStorage) do
		if container:IsShown() then
			container:SetScale(container:GetOption("Scale"))
		end
	end
end

--Set the number of items per row
function module:SetContainerRowSize()
	for name, container in pairs(containerStorage) do
		if container:IsShown() then
			container:SetAnchors()
		end
	end
end

function module:RefreshContainerAnchors()
	for name, container in pairs(containerStorage) do
		if container:IsShown() then
			container:SetAnchors()
		end
		if container.bagsBar and container.bagsBar:IsShown() then
			container:SetBagBarAnchors()
		end
	end
end

function module:RefreshContainerSlots()
	for name, container in pairs(containerStorage) do
		if container:IsShown() then
			for i = 1, container.NUM_BAG_IDS do
				local id = container.BAG_ID_LIST[i]
				for i = 1, #container.itemList[id] do
					container:SlotUpdate(container.itemList[id][i])
				end
			end
		end
	end
end

function module:RefreshContainerBackdrop()
	for name, container in pairs(containerStorage) do
		if container:IsShown() then
			-- Bag Backdrop
			module.bagBackdrop = {
				bgFile = Media:Fetch("background", container:GetOption("BackgroundTexture")),
				edgeFile = Media:Fetch("border", container:GetOption("BorderTexture")), 
				edgeSize = 10, insets = { left = 0, right = 0, top = 0, bottom = 0 }
			}
			-- Item Backdrop
			module.itemBackdrop = {
				bgFile = Media:Fetch("background", container:GetOption("BackgroundTexture")),
				edgeFile = Media:Fetch("border", container:GetOption("BorderTexture")),
				edgeSize = 15, insets = { left = 3, right = 3, top = 3, bottom = 3 },
			}
			
			module:ApplyBackdrop(container.background, module.bagBackdrop)
			for i = 1, container.NUM_BAG_IDS do
				local id = container.BAG_ID_LIST[i]
				for i = 1, #container.itemList[id] do
					module:ApplyBackdrop(self.itemList[id][i], module.itemBackdrop)
				end
			end
		end
	end
end

function module:RefreshColors()
	for name, container in pairs(containerStorage) do
		if container:IsShown() then
			container.background:SetBackdropColor(module:AlphaColor("Background"))
			container.background:SetBackdropBorderColor(module:AlphaColor("Border"))
		
			for i = 1, container.NUM_BAG_IDS do
				local id = container.BAG_ID_LIST[i]
				for i = 1, #container.itemList[id] do
					local itemSlot = self.itemList[id][i]
					itemSlot:SetBackdropColor(module:AlphaColor("Background"))
					itemSlot:SetBackdropBorderColor(module:AlphaColor("Border"))
				end
			end
		end
	end
end

function module:LoadOptions()
	local options = {
		Header = module:NewHeader(L["Bags_Name"], 1),
		Settings = module:NewRootGroup(L["Generic_Settings"], 2, nil, nil, {
			--TODO: Add option-based functions.
			Lock = module:NewToggle(L["Bags_Lock_Name"], L["Bags_Lock_Desc"], 1),
			ItemQuality = module:NewToggle(L["Bags_ShowItemQuality_Name"], L["Bags_ShowItemQuality_Desc"], 2, "RefreshContainerSlots"),
			ShowNew = module:NewToggle(L["Bags_ShowNewItemAnim_Name"], L["Bags_ShowNewItemAnim_Desc"], 3, "RefreshContainerSlots"),
			ShowQuest = module:NewToggle(L["Bags_ShowQuestHighlights_Name"], L["Bags_ShowQuestHighlights_Desc"], 4, "RefreshContainerSlots"),
			Header1 = module:NewHeader("",5),
			RowSize = module:NewSlider(L["Bags_ItemsPerRow_Name"], L["Bags_ItemsPerRow_Desc"], 6, 8, 32, 1, nil,  "RefreshContainerAnchors"),
			Scale = module:NewSlider(L["Generic_Scale"], L["Bags_Scale_Desc"], 7, 0.5, 2.5, 0.25, true, "SetContainerScale"),
			Padding = module:NewSlider(L["Generic_Padding"], L["Bags_Padding_Desc"], 8, 4, 24, 1, nil, "RefreshContainerAnchors"),
			Spacing = module:NewSlider(L["Generic_Spacing"], L["Bags_Spacing_Desc"], 9, 1, 15, 1, nil, "RefreshContainerAnchors"),
			
		}),
		Textures = module:NewGroup(L["Generic_Textures"], 3, nil, nil, {
			Background = module:NewHeader(L["Generic_Background"], 1),
			BackgroundTexture = module:NewTexBackground(L["Tooltip_BackgroundTex_Name"], L["Generic_BackgroundDesc"], 2, "RefreshContainerBackdrop", "double"),
			Border = module:NewHeader(L["Generic_Border"], 5),
			BorderTexture = module:NewTexBorder(L["Tooltip_BorderTex_Name"], L["Generic_BorderDesc"], 6, "RefreshContainerBackdrop", "double"),
			--aBorderSize = module:NewSlider(L["Tooltip_BorderSize_Name"], L["Tooltip_BorderSize_Desc"], 7, 1, 30, 1, nil, "RefreshContainerBackdrop", "double"),
		}),
	}
	return options
end

------------------------------------------------------
-- / FRAMEWORK FUNCTIONS / --
------------------------------------------------------
module.enableButton = true

function module:OnInitialize()
	LUI:RegisterModule(module)
	--Make elements share EnabledState with their modules.
	--TODO: Support for different enable states when we have GBank/Void support.
	for name, element in module:IterateModules() do
		element:SetEnabledState(module:IsEnabled())
	end
end

function module:OnEnable()
	db = module:GetDB()
	
	module.bagBackdrop = {
		bgFile = Media:Fetch("background", db.Textures.Background),
		edgeFile = Media:Fetch("border", db.Textures.Border), 
		edgeSize = 10, insets = { left = 0, right = 0, top = 0, bottom = 0 }
	}
	module.itemBackdrop = {
		bgFile = Media:Fetch("background", db.Textures.Background),
		edgeFile = Media:Fetch("border", db.Textures.Border),
		edgeSize = 15, insets = { left = 3, right = 3, top = 3, bottom = 3 },
	}
	
	for name, element in module:IterateModules() do
		module:EnableModule(name)
	end
	
	-- close bags before Enabling/Disabling the module
	CloseAllBags()
end

function module:OnDisable()
	CloseAllBags()
end