-- Clock Infotext

------------------------------------------------------
-- / SETUP AND LOCALS / --
------------------------------------------------------
local addonname, LUI = ...
local module = LUI:GetModule("Infotext")
local element = module:NewElement("Clock", "AceEvent-3.0", "AceHook-3.0")
local L = LUI.L
local db

-- local copies
local strtrim, strmatch = strtrim, strmatch
local gsub, format = gsub, format
local tonumber, date = tonumber, date
local GetNumSavedWorldBosses = GetNumSavedWorldBosses
local GetSavedWorldBossInfo = GetSavedWorldBossInfo
local GetSavedInstanceInfo = GetSavedInstanceInfo
local GetNumSavedInstances = GetNumSavedInstances
local TimeBreakDown = ChatFrame_TimeBreakDown
local GetInstanceInfo = GetInstanceInfo
local IsInInstance = IsInInstance
local GetGameTime = GetGameTime

-- local variables
local cvarLocal, cvarMilitary -- Cache containing CVars
local guildParty         -- If there's a G or not in the text
local instanceInfo       -- Any instance tag would go in this.
local invitesPending = false

-- constants
local GAMETIME_TOOLTIP_TOGGLE_CALENDAR = GAMETIME_TOOLTIP_TOGGLE_CALENDAR
local TIMEMANAGER_TOOLTIP_LOCALTIME = TIMEMANAGER_TOOLTIP_LOCALTIME
local TIMEMANAGER_TOOLTIP_REALMTIME = TIMEMANAGER_TOOLTIP_REALMTIME
local TIMEMANAGER_TITLE = TIMEMANAGER_TITLE
local TIMEMANAGER_PM = TIMEMANAGER_PM
local TIMEMANAGER_AM = TIMEMANAGER_AM
local CVAR_MILITARY = "timeMgrUseMilitaryTime"
local CVAR_LOCAL = "timeMgrUseLocalTime"

local CLOCK_UPDATE_TIME = 1

--Instance Difficulty constants
local TAG_GUILD_GROUP = " |cff66c7ffG|r"
local RAID_INFO_WORLD_BOSS = RAID_INFO_WORLD_BOSS
local INSTANCE_DIFFICULTY_FORMAT = {
	[1]  = "%d|cff00ff00 N|r",  -- 5man Normal
	[2]  = "%d|cffff0000 H|r",  -- 5man Heroic
	[3]  = "%d|cff00ff00 N|r",  -- 10man Normal (Legacy)
	[4]  = "%d|cff00ff00 N|r",  -- 25man Normal (Legacy)
	[5]  = "%d|cffff0000 H|r",  -- 10man Heroic (Legacy)
	[6]  = "%d|cffff0000 H|r",  -- 25man Heroic (Legacy)
	[7]  = "%d|cffaaaaaa L|r",  -- 25man Raid Finder
	[8]  = "%d|cffff0000 C|r",  -- 5man Challenge Mode
	[9]  = "%d|cff00ff00 N|r",  -- 40man Normal
	[10] = "Unused",            -- Unused (Do not break array)
	[11] = "%d|cffff0000 H|r",  -- 3man Heroic Scenario
	[12] = "%d|cff00ff00 N|r",  -- 3man Scenario
	[13] = "Unused",            -- Unused (Do not break array)
	[14] = "%d|cffffcc00 N|r",  -- Flexible Normal
	[15] = "%d|cffff0000 H|r",  -- Flexible Heroic
	[16] = "%d|cffff0000 M|r",  -- 20man Mythic
	[17] = "%d|cffaaaaaa L|r",  -- Flexible Raid Finder
	-- No clue what difficulty 18-20 are used for.
	[18] = "%d|cffaaaaaa E|r",   -- Event
	[19] = "%d|cffaaaaaa E|r",   -- Event
	[20] = "%d|cffaaaaaa E|r",   -- Event Scenario
}

-- TODO: Look into changing the Instance Difficulty Format using GetDifficultyInfo
-- name, groupType, isHeroic, isChallengeMode, displayHeroic, displayMythic, toggleDifficultyID = GetDifficultyInfo(id)

-- Defaults
element.defaults = {
	profile = {
		X = 1660,
		instanceDifficulty = true,
		showSavedRaids = true,
		showWorldBosses = true,
	},
}
------------------------------------------------------
-- / MODULE FUNCTIONS /
------------------------------------------------------

--First we format the secs into a day/hour/minute format. The leading space is necessary.
--Then we remove spaces followed by a 0, such as 0d or 0h. Those extra spaces are then trimmed.
local function formatTime(sec)
	local timeLeft = format(L["InfoClock_LockoutTimeLeft_Format"], TimeBreakDown(sec))
	timeLeft = gsub(timeLeft, L["InfoClock_LockoutTimeLeftGsub_Format"], "")
	return timeLeft:trim()
end

local function OneRaidCheck(bool)
	if not bool then
		GameTooltip:AddLine(" ")
		GameTooltip:AddLine(L["InfoClock_SavedRaids"])
	end
	return true
end

function element:UpdateInvites()
	invitesPending = (GameTimeFrame and GameTimeFrame.pendingCalendarInvites > 0) and true or false
end

function element:UpdateGuildParty()
	guildParty = InGuildParty() and TAG_GUILD_GROUP or nil
end

function element:UpdateInstanceInfo()
	local isInstance, instanceType = IsInInstance()
	if isInstance then
		local _, _, difficulty, _, _, _, _, _, groupSize = GetInstanceInfo()
		if instanceType == "raid" or instanceType == "party" then
			instanceInfo = format(INSTANCE_DIFFICULTY_FORMAT[difficulty], groupSize)
			return
		end
	end
	instanceInfo = nil
end

function element:UpdateCVar()
	cvarMilitary = GetCVarBool(CVAR_MILITARY)
	cvarLocal = GetCVarBool(CVAR_LOCAL)

	--HACK: Blizzard's TimeFrame checkboxes do not update when cvars are changed, so make sure they are up to date.
	TimeManagerMilitaryTimeCheck:SetChecked((cvarMilitary) and true or false)
	TimeManagerLocalTimeCheck:SetChecked((cvarLocal) and true or false)
	-- Only Refresh the options if the option panel is loaded.
	if element.RefreshOptionsPanel then 
		element:RefreshOptionsPanel() 
	end
	element:UpdateClock()
end

function element:GetTime(useLocal)
	local Hr, Min, PM
	if useLocal then
		Hr, Min = tonumber(date("%H")), date("%M")
	else
		Hr, Min = GetGameTime()
	end
	if not cvarMilitary then
		PM = (Hr >= 12) and TIMEMANAGER_PM or TIMEMANAGER_AM
		if Hr > 12 then
			Hr = Hr - 12
		elseif Hr == 0 then
			Hr = 12
		end
	end

	return format("%d:%.2d %s", Hr, Min, PM or ""):trim()
end

function element:UpdateClock()
	if invitesPending then
		element.text = L["InfoClock_InvitePending"]
	else
		local timeFormat = (db.instanceDifficulty and instanceInfo) and "%s (%s%s)" or "%s"
		element.text = format(timeFormat, element:GetTime(cvarLocal), instanceInfo or "", guildParty or "")
	end
	element:UpdateTooltip()
end

-- Click: Open Calendar Frame
-- RightClick: Open Time Manager
function element.OnClick(frame, button)
	if button == "RightButton" then
		TimeManager_Toggle()
	else
		GameTimeFrame:Click()
	end
end

function element.OnTooltipShow(GameTooltip)
	element:TooltipHeader(TIMEMANAGER_TITLE)

	--Display both set of times.
	GameTooltip:AddDoubleLine(cvarLocal and TIMEMANAGER_TOOLTIP_LOCALTIME or TIMEMANAGER_TOOLTIP_REALMTIME, element:GetTime(cvarLocal))
	GameTooltip:AddDoubleLine(cvarLocal and TIMEMANAGER_TOOLTIP_REALMTIME or TIMEMANAGER_TOOLTIP_LOCALTIME, element:GetTime(not cvarLocal))

	local oneraid -- Used so we dont display "Saved Raids:" unless you are saved to at least one.
	if db.showSavedRaids then
		for i = 1, GetNumSavedInstances() do
			local name, _, reset, difficulty, locked, extended, _, isRaid, maxPlayers, _, maxBosses, defeatedBosses = GetSavedInstanceInfo(i)
			if isRaid and (locked or extended) then
				local diff = (difficulty == 5 or difficulty == 6) and "H" or "N"
				local r, g, b = 1, 1, 1
				if extended then r, g, b = 0.5, 1, 0.5 end
				oneraid = OneRaidCheck(oneraid)
				GameTooltip:AddDoubleLine(format("%s |cffaaaaaa%s%s (%s/%s)", name, maxPlayers, diff, defeatedBosses, maxBosses), formatTime(reset), 1,1,1, r,g,b)
			end
		end
	end
	--Check for World Bosses toow
	if db.showWorldBosses then
		for i = 1, GetNumSavedWorldBosses() do
			local name, _, reset = GetSavedWorldBossInfo(i)
			oneraid = OneRaidCheck(oneraid)
			GameTooltip:AddDoubleLine(format("%s |cffaaaaaa(%s)", name, RAID_INFO_WORLD_BOSS), formatTime(reset), 1,1,1, 1,1,1)
		end
	end

	element:AddHint(GAMETIME_TOOLTIP_TOGGLE_CALENDAR, L["InfoClock_Hint_Right"])
end

------------------------------------------------------
-- / API FUNCTIONS / --
------------------------------------------------------
function element:LoadOptions()
	local function MilitaryTime(info, value)
		--Set
		if type(value) == "boolean" then
			SetCVar(CVAR_MILITARY, value and 1 or 0, true)
			element:UpdateCVar()
		--Get
		else
			return cvarMilitary
		end
	end
	local function LocalTime(info, value)
		--Set
		if type(value) == "boolean" then
			SetCVar(CVAR_LOCAL, value and 1 or 0, true)
			element:UpdateCVar()
		--Get
		else
			return cvarLocal
		end
	end
	local militaryMeta = { get = MilitaryTime, set = MilitaryTime }
	local localMeta = { get = LocalTime, set = LocalTime }
	
	local options = {
		setClock24h = element:NewToggle(TIMEMANAGER_24HOURMODE, nil, 1, militaryMeta, "normal"),
		setClockLocal = element:NewToggle(TIMEMANAGER_LOCALTIME, nil, 2, localMeta, "normal"),
		instanceDifficulty = element:NewToggle(L["InfoClock_InstanceDifficulty_Name"], L["InfoClock_InstanceDifficulty_Desc"], 3, "UpdateClock"),
		showSavedRaids = element:NewToggle(L["InfoClock_ShowSavedRaids_Name"], L["InfoClock_ShowSavedRaids_Desc"], 5, "UpdateTooltip"),
		showWorldBosses = element:NewToggle(L["InfoClock_ShowWorldBosses_Name"], L["InfoClock_ShowWorldBosses_Desc"], 6, "UpdateTooltip"),

	}
	return options
end

function element:OnCreate()
	db = element:GetDB()
	-- Update tags that can be found next to the clock.
	element:RegisterEvent("GUILD_PARTY_STATE_UPDATED", "UpdateGuildParty")
	element:RegisterEvent("PLAYER_DIFFICULTY_CHANGED", "UpdateInstanceInfo")
	element:RegisterEvent("INSTANCE_GROUP_SIZE_CHANGED", "UpdateInstanceInfo")
	element:RegisterEvent("PLAYER_ENTERING_WORLD", "UpdateInstanceInfo")
	element:UpdateGuildParty()
	element:UpdateInstanceInfo()

	-- Update cached CVar data.
	element:SecureHookScript(TimeManagerMilitaryTimeCheck, "OnClick", "UpdateCVar")
	element:SecureHookScript(TimeManagerLocalTimeCheck, "OnClick", "UpdateCVar")
	element:UpdateCVar()

	-- Update calendar invites
	element:RegisterEvent("CALENDAR_UPDATE_PENDING_INVITES", "UpdateInvites")
	element:UpdateInvites()

	element:AddUpdate("UpdateClock", CLOCK_UPDATE_TIME)
end